﻿#
# fiscalCodeObj.py
#
#

import  datetime

vowel_list = ['A', 'E', 'I', 'O', 'U']

month_codes = {1: 'A', 2: 'B', 3: 'C', 4: 'D', 5:'E', 6: 'H', 
				7: 'L', 8: 'M', 9: 'P', 10: 'R', 11: 'S', 12: 'T'}

transcode_even_chars = {'0':0, '1':1, '2':2, '3':3, '4':4, '5':5, '6':6, '7':7, '8':8, '9':9, 
						'A':0, 'B':1, 'C':2, 'D':3, 'E':4, 'F':5, 'G':6, 'H':7, 'I':8, 'J':9, 
						'K':10, 'L':11, 'M':12, 'N':13, 'O':14, 'P':15, 'Q':16, 'R':17, 'S':18,
						'T':19, 'U':20, 'V':21, 'W':22, 'X':23, 'Y':24, 'Z':25}

transcode_odd_chars = {'0':1, '1':0, '2':5, '3':7, '4':9, '5':13, '6':15, '7':17, '8':19, '9':21, 
						'A':1, 'B':0, 'C':5, 'D':7, 'E':9, 'F':13, 'G':15, 'H':17, 'I':19, 'J':21, 
						'K':2, 'L':4, 'M':18, 'N':20, 'O':11, 'P':3, 'Q':6, 'R':8, 'S':12,
						'T':14, 'U':16, 'V':10, 'W':22, 'X':25, 'Y':24, 'Z':23}

reminder = {0:'A', 1:'B', 2:'C', 3:'D', 4:'E', 5:'F', 6:'G', 7:'H', 8:'I', 9:'J', 10:'K', 11:'L',
			12:'M', 13:'N', 14:'O', 15:'P', 16:'Q', 17:'R', 18:'S', 19:'T', 20:'U', 21:'V', 22:'W', 23:'X', 24:'Y', 25:'Z' }
						
class fiscalCodeObj:

	def __init__(self, personal_data={}):
		self.personal_data = personal_data
		self.log_validation = ""
	

	def set_personal_data(self, pd):
		self.personal_data = pd
	
	def get_surname_code(self):
		surname = self.personal_data.get("surname", "").upper()
		consonant = ""
		vowel = ""
		for c in surname:
			if c in vowel_list:
				vowel += c
			else:
				consonant += c	
		result = consonant + vowel
		if len(result) < 3:
			result += 'X' * (3 - len(result))
		return result[:3]


	def get_name_code(self):
		surname = self.personal_data.get("name", "").upper()
		consonant = ""
		vowel = ""
		for c in surname:
			if c in vowel_list:
				vowel += c
			else:
				consonant += c	
		if len(consonant) > 3:
			return consonant[0] + consonant[2] + consonant[2]

		result = consonant + vowel
		if len(result) < 3:
			result += 'X' * (3 - len(result))
		return result[:3]

	def get_birthday_code(self):
		bday = self.personal_data.get("birthday")
		sex = self.personal_data.get("sex").upper()
		if not isinstance(bday, datetime.date):
			#error: maybe raise exception?
			return
		_year = str(bday.year % 100)
		if len(_year) == 1:
			_year = '0' + _year
		_month = month_codes.get(bday.month)
		if sex == 'F':
			_day = str(bday.day + 40)
		else:
			_day = str(bday.day)
			if len(_day) == 1:
				_day = '0' + _day
		return _year + _month + _day

	def get_birthplace_code(self):
		#not impelented
		return ''
		
		
	def get_control_code(self, fc=None):
		if fc is None:
			#create fiscal code based on personal data
			fc = self.get_surname_code() + self.get_name_code() + self.get_birthday_code() + self.get_birthplace_code()
			if len(calculated_fc) != 15:
				#maybe birthplace part is not implemented 
				#so we can't calculate the control code
				return None
		total = 0
		for i, c in enumerate(fc):
			#be careful that ad odd position in a python string
			#is an even position in realilty because index starts
			#from zero 
			if i % 2:
				total += transcode_even_chars[c]
			else:	
				total += transcode_odd_chars[c]
		return reminder[total % 26 ]
	
	def validate_fiscalcode(self, fc):
		fc = fc.upper()
		self.log_validation = ''
		if len(fc) != 16:
			self.log_validation += fc + ' invalid length: ' + str(len(fc)) + ' instead of 16 \n'
			return False
		s_code = self.get_surname_code()
		if s_code != fc[:3]:
			self.log_validation += 'Surname part: get ' + fc[:3] + ' instead of ' + s_code + '\n'
			return False
		else:
			self.log_validation += 'Surname part: ' + s_code + ' Ok!\n'
		
		n_code = self.get_name_code()
		if n_code != fc[3:6]:
			self.log_validation += 'Name part: get ' + fc[3:6] + ' instead of ' + n_code + '\n'
			return False
		else:
			self.log_validation += 'Name part: ' + n_code + ' Ok!\n'
		
		b_code = self.get_birthday_code()
		if b_code != fc[6:11]:
			self.log_validation += 'Birthday part: get ' + fc[6:11] + ' instead of ' + b_code + '\n'
			return False
		else:
			self.log_validation += 'Birthday part: ' + b_code + ' Ok!\n'

		bp_code = self.get_birthplace_code()
		if len(bp_code) > 0 :
			if bp_code != fc[11:15]:
				self.log_validation += 'Birthplace part: get ' + fc[11:15] + ' instead of ' + bp_code +  '\n'
				return False
			else:
				self.log_validation += 'Birthplace part: ' + bp_code + ' Ok!\n'
		else:
			self.log_validation += 'Birthplace part: skipped \n'

		c_code = self.get_control_code(fc[:15])
		if c_code != fc[15]:
			self.log_validation += 'Control Code part: get ' + fc[15] + ' instead of ' + c_code +  '\n'
			return False
		else:
			self.log_validation += 'Control Code part: ' + c_code + ' Ok!\n'
		return True


	
	def get_validation_log(self):
		return self.log_validation

		
if __name__ == '__main__':
	#test
	_s = raw_input('Your surname: ')
	_n = raw_input('Your name: ')
	_sx = raw_input('Male or Female (M/F): ')
	_bd = raw_input('Your birthdate (dd/mm/yyyy): ')
	_fc = raw_input('Your fiscal code: ')
	
	p_data = {'surname': _s,
			   'name': _n,
			   'birthday': datetime.date(int(_bd[-4:]), int(_bd[3:5]), int(_bd[:2])),
			   'sex': _sx,
			   'birthplace': ''}
	fc = fiscalCodeObj(p_data)
	fc.validate_fiscalcode(_fc)
	print fc.get_validation_log()
	
